function GpsManager(interactiveMap) {
  this.map = interactiveMap.map;
  this.control = new maplibregl.GeolocateControl({
    positionOptions: {
      enableHighAccuracy: true,
      maximumAge: 3000,
    },
    trackUserLocation: true,
  });
  this.controlAdded = false;
  this.interactiveMap = interactiveMap;
  this.gps = null; // Boolean which declares whether gps is available or not so we don't have to check against the API everytime
  this.location = null; // Array with Position data of the Last Position we retrieved
  this.bearing = null;
  this.lockViewToPosition = true; // Whether the view should be locked when the current Location is shown.
  this.id = null; // ID of the process that follow the Location
  this.userPositionMarker = null; // Marker that displays the user Position
  this.point_geom = null; // Geomatry of the exact Point of the user
  this.circle = null; // Geometry of the accuracy of the user position
  this.options = {
    enableHighAccuracy: true,
    maximumAge: 0,
  };
  this.checkGps(); // This function will set the value of "this.gps" it check gps availability asynchronious
  // Be carefull we will not know if we have GPS or not directly after this constructor
  // because the validation is done asynchroniously.
}

GpsManager.prototype.constructor = GpsManager;

GpsManager.prototype.loadingGps = function() {
  // Returns a boolean so you can check whether this Manager is finished loading Gps
  if (this.gps == null) return true;
  else return false;
};

GpsManager.prototype.getLocation = function() {
  var caller = this;
  return new Promise((resolve, reject) => {
    if (navigator.geolocation) {
      navigator.geolocation.getCurrentPosition((position) => {
        caller.toggleGpsLocator(true);
        caller.location = [position.coords.longitude, position.coords.latitude];
        caller.accuracy = position.coords.accuracy;

        caller.gps = true;
        caller.enableGpsFeatures();
        resolve(position);
      }, reject);
    } else {
      reject;
    }
  });
};

GpsManager.prototype.checkGps = function() {
  if (navigator.geolocation) {
    var caller = this;
    navigator.geolocation.getCurrentPosition(
      function(position) {
        caller.toggleGpsLocator(true);
        caller.location = [position.coords.longitude, position.coords.latitude];
        caller.accuracy = position.coords.accuracy;

        caller.gps = true;
        caller.enableGpsFeatures();
      },
      function(error) {
        caller.gps = false;
        caller.toggleGpsLocator(false);
        caller.toggleGpsWarning();
        caller.disableGpsFeatures();
      },
      {
        enableHighAccuracy: true,
        maximumAge: 0,
      }
    );
  } else {
    this.gps = false;
    this.toggleGpsLocator(false);
    this.toggleGpsWarning();
    this.disableGpsFeatures();
  }
};

// The GpsManager can call the predefined Functions of the current module
// If it finishes too fast with fetching the position it can cause troubles thats why we add a timeout
GpsManager.prototype.enableGpsFeatures = function() {
  window.setTimeout(
    $.proxy(function() {
      this.interactiveMap.module.enableGps();
    }, this),
    100
  );
  //setTimeout(this.interactiveMap.module.enableGps(), 10000);
};
GpsManager.prototype.disableGpsFeatures = function() {
  window.setTimeout(
    $.proxy(function() {
      this.interactiveMap.module.disableGps();
    }, this),
    100
  );
};
/**
 * Toggles the Map Feature (GpsLocation)
 * It's a button on the map to display your own Position
 * @param visible - Boolean whether GPS is available or not
 **/
GpsManager.prototype.toggleGpsLocator = function(visible) {
  if (visible) {
    if (!this.controlAdded) {
      this.map.addControl(this.control, "bottom-left");
      this.controlAdded = true;
    }
    $("#start-navigation > a").attr("href", "/route/start/foot/gps;");
  } else {
    if (this.controlAdded) {
      this.map.removeControl(this.control);
      this.controlAdded = false;
    }
    $("#start-navigation > a").attr("href", "/route/start/foot");
  }
};
/**
 * If the retrieval of GPS Position fails on mobile devices we will show a small warning for a
 * period of time.
 **/
GpsManager.prototype.toggleGpsWarning = function() {
  $("#gps-error").addClass("visible-xs");
  $("#gps-error").removeClass("hidden");
  setTimeout(function() {
    $("#gps-error").addClass("hidden");
    $("#gps-error").removeClass("visible-xs");
  }, 5000);
};

GpsManager.prototype.stopWatch = function() {
  if (typeof this.followId != "undefined") {
    navigator.geolocation.clearWatch(this.followId);
    this.followId = undefined;
    console.log("Watch stopped");
  }
};

GpsManager.prototype.watchPosition = function(callback, options) {
  if (typeof options == "undefined") {
    var options = {
      enableHighAccuracy: true,
      maximumAge: 3000,
    };
  }
  if (typeof callback != "function") return;
  this.followId = navigator.geolocation.watchPosition(
    $.proxy(function(position) {
      // We have a new Position
      var long = parseFloat(position.coords.longitude);
      var lat = parseFloat(position.coords.latitude);
      this.location[0] = long;
      this.location[1] = lat;
      this.accuracy = 20; // parseFloat(position.coords.accuracy); TODO REMOVE
      this.timestamp = Math.floor(position.timestamp / 1000);
      // We can calculate the possible Heading if this is not the first Position we retrieve
      if (
        typeof this.bearingPos == "object" &&
        this.bearingPos.length == 2 &&
        this.getDistance(this.bearingPos, this.location) > this.accuracy
      ) {
        this.bearing = this.getBearing(this.bearingPos, this.location);
        this.bearingPos = [long, lat];
      } else {
        this.bearingPos = [long, lat];
        this.bearing = null;
      }
      callback(position);
    }, this),
    function(error) {
      // Follow Location couldn't be started. Abort now
      deinitAssistent();
    },
    options
  );
};

GpsManager.prototype.getBearing = function(p1, p2) {
  var point1 = turf.point(p1);
  var point2 = turf.point(p2);
  var bearing = turf.bearing(point1, point2);
  if (bearing < 0) {
    bearing += 360;
  }
  return bearing;
};

GpsManager.prototype.toRadians = function(angle) {
  return angle * (Math.PI / 180);
};

GpsManager.prototype.toDegrees = function(radians) {
  return (radians * 180) / Math.PI;
};

GpsManager.prototype.getDistance = function(p1, p2) {
  var point1 = turf.point(p1);
  var point2 = turf.point(p2);
  var dist = turf.distance(point1, point2, { units: "kilometers" }) * 1000;
  return dist;
};
